;;;; sequences-utils.scm  -*- Scheme -*-
;;;; Kon Lovett, Oct '22

;;Issues
;;

(module (sequences utils)

(;export
  ;
  foldl->alist
  ;
  ->list ->vector ->string
  ;
  merge merge!
  sorted?
  sort sort!
  unique
  ;
  histogram
  ;
  randoms
  sample)

(import scheme
  (chicken base)
  (chicken type)
  (chicken fixnum)
  (chicken random)
  (only (srfi 1) list-copy reverse!)
  (prefix (chicken sort) chicken:)
  (prefix sequences seq:))

;;;

(include "sequences-utils-types")

(: foldl->alist (seq (* 'a -> 'a) 'a #!optional binary-equality --> (list-of (pair * 'a))))

(: ->list (seq --> list))
(: ->vector (seq --> vector))
(: ->string (seq --> string))

(: merge! (seq seq binary-predicate -> seq))
(: merge (seq seq binary-predicate --> seq))

(: sorted? (seq binary-predicate --> boolean))

(: sort! (seq binary-predicate -> seq))
(: sort (seq binary-predicate --> seq))

(: histogram (seq #!optional binary-equality --> alist))

;;

;miscellaneous

(define (foldl->alist seq next seed #!optional (eqal? equal=?))
  (define (kons bins value)
    (let ((cur (alist-ref value bins eqal? seed)))
      (alist-update! value (next value cur) bins eqal?)) )
  (seq:foldl kons '() seq) )

;basic coercions

(define (->list x) (if (list? x) x (seq:coerce '() x)))
(define (->vector x) (if (vector? x) x (seq:coerce #() x)))
(define (->string x) (if (string? x) x (seq:coerce "" x)))

;expose (chicken sort)

(define (merge! a b less?)
  (let ((rs (chicken:merge! (->list a) (->list b) less?)))
    (if (list? a) rs (seq:coerce a rs)) ) )

(define (merge a b less?)
  (let ((rs (chicken:merge (->list a) (->list b) less?)))
    (if (list? a) rs (seq:coerce a rs)) ) )

(define (sorted? seq less?)
  (let ((ss (if (vector? seq) seq (->list seq))))
    (chicken:sorted? ss less?) ) )

(define (sort! seq less?)
  ;FIXME fill-in original from sorted-list
  (let* ((ss (if (vector? seq) seq (->list seq)))
         (rs (chicken:sort! ss less?)) )
    (if (or (list? seq) (vector? seq)) rs (seq:coerce seq rs)) ) )

(define (sort seq less?)
  (define (*->list x) (if (list? x) (list-copy x) (->list x)))
  (seq:coerce seq (chicken:sort! (*->list seq) less?)) )

(define (unique seq #!optional (eqal? equal?))
  (assert (seq:sequence? seq))
  #; ;assumed, cannot enforce
  (assert (sorted? seq lessp))
  (assert (procedure? eqal?))
  (if (seq:empty? seq) seq
      (let ((cur (seq:peek seq)))
        (let loop ((seq (seq:pop seq)) (os (list cur)) (prv cur))
          (if (seq:empty? seq) (seq:coerce seq (reverse! os))
              (let ((cur (seq:peek seq)) (rst (seq:pop seq)))
                (if (eqal? prv cur)
                    (loop rst os prv)
                    (loop rst (cons cur os) cur) ) ) ) ) ) ) )

;really simple stats

(define (histogram seq #!optional (eqal? equal=?))
  (foldl->alist seq (lambda (_ occurs) (add1 occurs)) 0 eqal?) )

;high is end
(define (randoms cnt #!optional (high most-positive-fixnum) (low 0))
  (assert (fixnum? cnt))
  (assert (integer? high))
  (assert (integer? low))
  (let ((lim (- high low)))
    (let loop ((cnt cnt) (ls '()))
      (if (fx<= cnt 0) ls
          (loop (fx- cnt 1) (cons (+ (pseudo-random-integer lim) low) ls)) ) ) ) )

(define (sample seq . rest)
  (assert (seq:sequence? seq))
  (let* ((siz (seq:size seq))
         (cnt (optional rest (pseudo-random-integer siz))) )
    #; ;assumed
    (assert (fixnum? siz))
    (assert (fixnum? cnt))
    (assert (fx<= cnt siz))
    ;use of a sorted set of indices allows only forward input motion
    (let ((is (unique (chicken:sort! (randoms cnt siz) fx<) fx=)))
      (define (linear-sample)
        (let loop ((seq seq) (is is) (i 0) (os '()))
          (cond ((or (null? is) (seq:empty? seq))
                  os)
                ((fx= i (car is))
                  (loop (seq:pop seq) (cdr is) (fx+ i 1) (cons (seq:peek seq) os)))
                (else
                  (loop (seq:pop seq) is (fx+ i 1) os)) ) ) )
      (define (random-sample)
        (let loop ((iter (seq:iterator seq)) (is is) (i 0) (os '()))
          (cond ((or (null? is) (seq:at-end? iter))
                  os)
                ((fx= i (car is))
                  (let ((elm (seq:elt seq (seq:index iter))))
                    (loop (seq:advance! iter) (cdr is) (fx+ i 1) (cons elm os))))
                (else
                  (loop (seq:advance! iter) is (fx+ i 1) os)) ) ) )
      (let ((os (if (seq:linear-sequence? seq) (linear-sample) (random-sample))))
          (seq:coerce seq (reverse! os)) ) ) ) )

) ;module (sequences utils)
